from dataclasses import dataclass
from datetime import date, datetime
from functools import reduce
from typing import Any, Callable, Generic, Literal, Protocol, Self, Tuple, Type, TypeVar
from uuid import UUID

from click import DateTime
from pydantic import BaseModel
from sqlalchemy import Select, delete, select, update
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

IDENTIFIABLE_SHEMA_T = TypeVar("IDENTIFIABLE_SHEMA_T")


class Identifiable(BaseModel, Generic[IDENTIFIABLE_SHEMA_T]):
    id: UUID
    data: IDENTIFIABLE_SHEMA_T


T_BASEMODEL = TypeVar("T_BASEMODEL", bound=BaseModel)


class Base(DeclarativeBase): ...


class DatedEntity(Base):
    __abstract__ = True

    created_at: Mapped[datetime] = mapped_column(nullable=False)
    updated_at: Mapped[datetime] = mapped_column(nullable=False)


class IdentifiableEntity(Base):
    __abstract__ = True

    id: Mapped[UUID] = mapped_column(primary_key=True)

    @classmethod
    def from_identifiable(cls, model: Identifiable[T_BASEMODEL]) -> "Self":
        return cls(**{"id": model.id, **model.data.model_dump()})

    @classmethod
    def from_basemodel(cls, mutation: T_BASEMODEL) -> "Self":
        return cls(**mutation.model_dump())

    def to_identifiable(self, MODEL: Type[T_BASEMODEL]) -> Identifiable[T_BASEMODEL]:

        return Identifiable[MODEL].model_validate(  # type: ignore[valid-type]
            {"id": self.id, "data": self.__dict__}
        )

    def to_basemodel(self, model: Type[T_BASEMODEL]) -> T_BASEMODEL:
        return model.model_validate(self.__dict__)


IDENTIFIABLE_T = TypeVar("IDENTIFIABLE_T", bound=IdentifiableEntity)


class CRUDOperations(Generic[IDENTIFIABLE_T]):

    def __init__(
        self,
        entity_type: Type[IDENTIFIABLE_T],
        session_provider: Callable[[], AsyncSession],
    ) -> None:
        self.entity_type = entity_type
        self.session_provider = session_provider

    @property
    def session(self) -> AsyncSession:
        return self.session_provider()

    async def create(self, entity: IDENTIFIABLE_T) -> None:
        self.session.add(entity)
        await self.session.flush()
        await self.session.refresh(entity)

    async def get(self, id: UUID) -> IDENTIFIABLE_T:
        return await self.session.get_one(self.entity_type, id)

    async def update(self, entity: IDENTIFIABLE_T) -> None:
        await self.session.merge(entity)

    async def delete(self, entity: IDENTIFIABLE_T) -> None:
        await self.session.delete(entity)

    async def delete_by_id(self, id: UUID) -> None:
        await self.session.execute(
            delete(self.entity_type).where(self.entity_type.id == id)
        )

    async def update_by_id(self, id: UUID, entity: IDENTIFIABLE_T) -> None:
        await self.session.execute(
            update(self.entity_type)
            .where(self.entity_type.id == id)
            .values(entity.__dict__)
        )


QUERY_ENTITY_T = TypeVar("QUERY_ENTITY_T", bound=Base)
QUERY_FILTER_T = TypeVar("QUERY_FILTER_T")


class QueryInjector(Protocol):

    def inject(self, query: Select[Tuple[Any]], filter: Any) -> Select[Tuple[Any]]: ...


class QueryOperations(Generic[QUERY_FILTER_T, QUERY_ENTITY_T]):

    def __init__(
        self,
        entity_type: Type[QUERY_ENTITY_T],
        session_provider: Callable[[], AsyncSession],
        filters_functions: list[QueryInjector],
    ) -> None:
        self.entity_type: type[QUERY_ENTITY_T] = entity_type
        self.session_provider = session_provider
        self.filters_functions = filters_functions

    @property
    def session(self) -> AsyncSession:
        return self.session_provider()

    async def query(self, filter: QUERY_FILTER_T) -> list[QUERY_ENTITY_T]:
        return [
            e
            for e in (
                await self.session.execute(
                    self.generate_filtered_query(filter, select(self.entity_type))
                )
            ).scalars()
        ]

    def generate_filtered_query(
        self, filter: QUERY_FILTER_T, select_query: Select[Tuple[QUERY_ENTITY_T]]
    ) -> Select[Tuple[QUERY_ENTITY_T]]:
        return reduce(
            lambda query, filter_function: filter_function.inject(query, filter),
            self.filters_functions,
            select_query,
        )


# region PaginatedFilter
@dataclass
class PaginatedFilter:
    offset: int = 0
    limit: int = 10


class PaginatedQueryInjector(QueryInjector):
    def inject(
        self,
        query: Select[Tuple[Base]],
        filter: PaginatedFilter,
    ) -> Select[Tuple[Base]]:
        return query.offset(filter.offset).limit(filter.limit)


# endregion


# region Criteria


@dataclass
class StringCriteria:
    value: str
    is_exact: bool
    case_sensitive: bool


@dataclass
class DateCriteria:
    value: date
    op: Literal["eq", "gt", "lt", "gte", "lte"]


@dataclass
class DatetimeCriteria:
    value: datetime
    op: Literal["eq", "gt", "lt", "gte", "lte"]


class CriteriaBasedAttributeQueryInjector(QueryInjector):

    def __init__(self, entity_type: Type[Base]) -> None:
        self.entity_type = entity_type

    def inject(self, query: Select[Tuple[Base]], filter: Any) -> Select[Tuple[Base]]:

        attrs = filter.__dict__

        for field_name, value in attrs.items():

            if isinstance(value, (DateCriteria, DatetimeCriteria)):
                value = getattr(filter, field_name)

                entity_field = getattr(self.entity_type, field_name)

                op_mapping = {
                    "eq": entity_field == value.value,
                    "gt": entity_field > value.value,
                    "lt": entity_field < value.value,
                    "gte": entity_field >= value.value,
                    "lte": entity_field <= value.value,
                }

                query = query.filter(op_mapping[value.op])
            elif isinstance(value, StringCriteria):
                value = getattr(filter, field_name)

                entity_field = getattr(self.entity_type, field_name)

                if value.is_exact:
                    query = query.filter(entity_field == value.value)
                else:
                    query = query.filter(entity_field.contains(value.value))

        return query


# endregion



import importlib

import click
import uvicorn

from jararaca.messagebus.worker import (
    AioPikaConfig,
    AioPikaMicroserviceProvider,
    create_messagebus_worker,
)
from jararaca.microservice import Microservice
from jararaca.presentation.server import create_http_server


def find_app_by_module_path(
    module_path: str,
) -> Microservice:
    if ":" not in module_path:
        raise ValueError("'%s' is not a valid module path" % module_path)

    module_name, app = module_path.rsplit(":", 1)

    try:
        module = importlib.import_module(module_name)
    except ImportError:
        raise ValueError("App module not found")

    if not hasattr(module, app):
        raise ValueError("module %s has no attribute %s" % (module, app))

    app = getattr(module, app)

    if not isinstance(app, Microservice):
        raise ValueError("App must be an instance of App")

    return app


@click.group()
def cli() -> None:
    pass


@cli.command()
@click.argument(
    "app_path",
    type=str,
)
def worker(app_path: str) -> None:

    app = find_app_by_module_path(app_path)

    create_messagebus_worker(app)


@cli.command()
@click.argument(
    "app_path",
    type=str,
)
@click.option(
    "--host",
    type=str,
    default="0.0.0.0",
)
@click.option(
    "--port",
    type=int,
    default=8000,
)
def server(app_path: str, host: str, port: int) -> None:

    app = find_app_by_module_path(app_path)

    asgi_app = create_http_server(app)

    uvicorn.run(asgi_app, host=host, port=port)



import asyncio
import logging
from dataclasses import dataclass
from typing import Any

import aio_pika
import uvloop

from jararaca.di import Container
from jararaca.messagebus.decorators import MESSAGEBUS_INCOMING_MAP, MessageBusController
from jararaca.microservice import Microservice


@dataclass
class AioPikaConfig:
    url: str
    queue: str
    exchange: str


class AioPikaMicroserviceProvider:
    def __init__(self, config: AioPikaConfig, incoming_map: MESSAGEBUS_INCOMING_MAP):
        self.config = config
        self.incoming_map = incoming_map

    def start_consumer(self) -> None:

        async def consume() -> None:
            connection = await aio_pika.connect_robust(self.config.url)

            channel = await connection.channel()

            topic_exchange = await channel.declare_exchange(
                "%s_topic" % self.config.exchange, type="topic"
            )

            queue = await channel.declare_queue(self.config.queue)

            topics = [*self.incoming_map.keys()]

            for topic in topics:
                await queue.bind(topic_exchange, routing_key=topic)

            await queue.consume(self.on_message)

            try:
                # Wait until terminate
                await asyncio.Future()
            finally:
                await connection.close()

            # async with queue.iterator() as queue_iter:

            #     async for message in queue_iter:

            #         async with message.process():
            #             topic = message.routing_key

            #             handler = self.incoming_map.get(topic)

            #             if handler is None:
            #                 logging.warning("No handler found for topic '%s'" % topic)
            #                 continue

            #             try:
            #                 await handler(message)
            #             except:
            #                 logging.exception("deu pau")
            #             finally:
            #                 await message.ack()

        with asyncio.Runner(loop_factory=uvloop.new_event_loop) as runner:
            runner.run(consume())

    async def on_message(self, message: aio_pika.abc.AbstractIncomingMessage) -> None:
        topic = message.routing_key

        if topic is None:
            logging.warning("No topic found for message")

            return

        handler = self.incoming_map.get(topic)

        if handler is None:
            logging.warning("No handler found for topic '%s'" % topic)
            return

        try:
            await handler(message)  # type: ignore[arg-type]  # TODO: Transform message from aio_pika to a more generic message
        except:
            logging.exception("deu pau")
        finally:
            await message.ack()


def create_messagebus_worker(app: Microservice) -> None:
    container = Container(app)

    combined_messagebus_incoming_map: MESSAGEBUS_INCOMING_MAP = {}

    for instance_type in app.controllers:
        controller = MessageBusController.get_messagebus(instance_type)

        if controller is None:
            continue

        instance: Any = container.get_by_type(instance_type)

        factory = controller.get_messagebus_factory()(instance)

        for topic, handler in factory.items():
            if topic in combined_messagebus_incoming_map:
                logging.warning(
                    "Topic '%s' already registered by another controller" % topic
                )
            combined_messagebus_incoming_map[topic] = handler

        factory

    AioPikaMicroserviceProvider(
        config=AioPikaConfig(
            url="amqp://guest:guest@localhost/",
            exchange="test_exchange",
            queue="test_queue",
        ),
        incoming_map=combined_messagebus_incoming_map,
    ).start_consumer()



from dataclasses import dataclass, field
from typing import Any, AsyncContextManager, Callable, Protocol

from jararaca.core.providers import ProviderSpec


class AppInterceptor(Protocol):

    def intercept(self) -> AsyncContextManager[None]: ...


@dataclass
class Microservice:
    providers: list[type[Any] | ProviderSpec] = field(default_factory=list)
    controllers: list[type] = field(default_factory=list)
    interceptors: list[AppInterceptor] = field(default_factory=list)



from typing import Any, AsyncGenerator, Callable

from fastapi import Depends, FastAPI

from jararaca.di import Container
from jararaca.microservice import Microservice
from jararaca.presentation.decorators import RestController


class UowDependency:

    def __init__(self, app: Microservice):
        self.app = app

    async def __call__(self) -> AsyncGenerator[None, None]:

        ctxs = [interceptor.intercept() for interceptor in self.app.interceptors]

        for ctx in ctxs:
            await ctx.__aenter__()

        yield None

        for ctx in reversed(ctxs):
            await ctx.__aexit__(None, None, None)


def create_http_server(
    app: Microservice,
    factory: Callable[[], FastAPI] | None = None,
) -> FastAPI:
    container = Container(app)

    fastapi_app = factory() if factory is not None else FastAPI()

    fastapi_app.router.dependencies.append(Depends(UowDependency(app)))

    for controller_t in app.controllers:
        controller = RestController.get_controller(controller_t)

        if controller is None:
            continue

        instance: Any = container.get_by_type(controller_t)

        router = controller.get_router_factory()(instance)

        fastapi_app.include_router(router)

    return fastapi_app



from .microservice import Microservice
from .persistence.interceptors.aiosqa_interceptor import (
    AIOSqlAlchemySessionInterceptor,
    AIOSQAConfig,
)
from .core.providers import ProviderSpec, Token
from .presentation.server import create_http_server
from .messagebus.worker import create_messagebus_worker
from .di import Container


__all__ = [
    "Microservice",
    "ProviderSpec",
    "Token",
    "AIOSqlAlchemySessionInterceptor",
    "AIOSQAConfig",
    "create_http_server",
    "create_messagebus_worker",
    "Container",
]
